#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <strings.h>
#include <math.h>
#include <ctype.h>
#include <dlfcn.h>
#include <errno.h>
#include "libswitch.h"
#include "libsfp.h"
#include "libdb.h" // x10-db

#define HW_HANDLE(lsw) LIBSW_HW_HANDLE(lsw)
#define LIBSFP_HANDLE(lsw) ((libsfp_t *)((lsw)->sfp.handle))

#define LIBSW_DEBUG_BASE_LEVEL 1

#ifdef LIBSW_DEBUG
#define LIBSWDEBUGL(level, format, ...) \
    debug_print(&(lsw->log), LIBSW_DEBUG_BASE_LEVEL+level, \
    "%s: "format"\n",__FUNCTION__, ##__VA_ARGS__)
#else
#define LIBSWDEBUGL(level, format, ...)
#endif

#define LIBSWDEBUG(format, ...) LIBSWDEBUGL(0, format, ##__VA_ARGS__)
#define LIBSWERR(format, ...)  LIBSWDEBUGL(0, "ERR: "format, ##__VA_ARGS__)



void *libsw_get_next_obj_iface(void *lsw, void *itr);
void *libsw_get_next_obj_vlan2(void *lsw, void *itr);
void *libsw_get_next_obj_lag(void *lsw, void *itr);
void *libsw_get_next_obj_vmt(void *lsw, void *itr);

#include "libsw_drvfn.h"

char default_hwname[]="x10";

int libsw_load_hwlib(libsw_t *lsw, char *name)
{
  uint16_t i;
  void **p;
  char fn[128];
  char funname[64];

  snprintf(fn, sizeof(fn), LIBSW_MODULE_NAME_TEMPLATE, name);

  LIBSWDEBUG("libfile:'%s'",fn);

  lsw->dllhandle = dlopen(fn, RTLD_NOW);

  if (!(lsw->dllhandle)) {
    LIBSWERR("Can't open module '%s' - '%s'", fn, dlerror());
    return LIBSW_ERROR;
  }

  p = (void *)&(lsw->hw->init);
  lsw->hw->name = name;
  for (i=0; i<(sizeof(hw_func_name)/sizeof(char*)); ++i) {
    snprintf(funname, sizeof(funname), "fujitsu_%s", hw_func_name[i]);
    p[i] = dlsym(lsw->dllhandle, funname);
    LIBSWDEBUGL(2, "symbol %i '%s' %p ", i, funname, p[i]);
    if (!p[i]) {
      LIBSWERR("Can't resolve symbol '%s' in module '%s' - '%s'",
             funname, fn, dlerror());
      return LIBSW_ERROR;
    }
  }

  return 0;
}

int libsw_s2n10(const char *s, uint16_t *n)
{
  const char *ss=s;
  char *p;

  while (*ss) {
    if (!isdigit(*ss))
      return LIBSW_ERROR;
    ss++;
  }

  errno=0;
  (*n) = (uint16_t)strtoul(s, &p, 10);

  if (errno)
    return 1;

  if (*p)
    return 1;

  return 0;
}

void libsw_set_debug(libsw_t *lsw, int l, FILE *f)
{
  lsw->log.level = l;
  if (f)
   lsw->log.file = f;
}

int libsfp_read_sfp_regs(void *udata, uint8_t i2caddr, uint16_t start, uint16_t count, void *data)
{
  libsw_t *lsw = (libsw_t *)udata;
  return lsw->hw->read_sfp_regs(HW_HANDLE(lsw), lsw->sfp.iface, i2caddr, start, count, data);
}

int libsw_init(libsw_t **sw, char *hwname, int dgblevel)
{
  int r=0;

  /* create library handle */
  *sw = malloc(sizeof(libsw_t));
  if (!(*sw))
      return LIBSW_ERROR;
  memset(*sw, 0, sizeof(libsw_t));

  /* Create debug log buffer */
  if (debug_init( &((*sw)->log) , dgblevel)) {
    r = LIBSW_ERROR;
    goto exit;
  }  
  (*sw)->prm.log = &((*sw)->log);

  if (!hwname)
    hwname = getenv("LIBSW_DEFAULT_DRIVER");

  debug_print( &((*sw)->log),
               LIBSW_DEBUG_BASE_LEVEL,
               "%s: lev:%u hw:'%s'\n", __FUNCTION__, dgblevel, hwname );

  /* Create Buffer for hw driver object */
  (*sw)->hw = malloc(sizeof(switch_hw_t));
  if (!((*sw)->hw)) {
    r = LIBSW_ERROR;
    goto exit;
  }


  /* Load hw driver library */
  if (libsw_load_hwlib(*sw, hwname?hwname:default_hwname)) {
    r = LIBSW_ERROR;
    goto exit;
  }

  /* Initialize libsfp */
  libsfp_t *sfp;
  if (libsfp_init(&sfp)) {
    r = LIBSW_ERROR;
    goto exit;
  }

  (*sw)->sfp.handle = sfp;
  libsfp_set_readreg_callback(LIBSFP_HANDLE((*sw)), libsfp_read_sfp_regs);
  libsfp_set_user_data(LIBSFP_HANDLE((*sw)), (*sw));
  (*sw)->prm.sfp = &((*sw)->sfp);

  /* Initialize DB API */
  debug_print( &((*sw)->log), LIBSW_DEBUG_BASE_LEVEL, "%s: Connect to the DBA: %s\n", __FUNCTION__, "default" );
  (*sw)->dbapi = libsw_get_db_api( NULL, NULL );
  if ( ! (*sw)->dbapi ) {
    debug_print( &((*sw)->log), LIBSW_DEBUG_BASE_LEVEL, "%s: Unable to connect to the DBA: %s\n", __FUNCTION__, "default" );
    r = LIBSW_ERROR;
    goto exit;
  }
  (*sw)->prm.dbapi = (*sw)->dbapi;
  
  /* Initialize MAC DB */
  debug_print( &((*sw)->log), LIBSW_DEBUG_BASE_LEVEL, "%s: Connect to the DB: %s\n", __FUNCTION__, "mac-table" );
  (*sw)->mac_db = NULL;
  (*sw)->mac_db = (*sw)->dbapi->open( (*sw)->dbapi,
                                      "mac-table",
                                      1024,
                                      sizeof( mac_table_t ),
                                      0,
                                      NULL, NULL );
  if ( ! (*sw)->mac_db ) {
    debug_print( &((*sw)->log), LIBSW_DEBUG_BASE_LEVEL, "%s: Unable to connect to the DB: %s\n", __FUNCTION__, "mac-table" );
    r = LIBSW_ERROR;
    goto exit;
  }

  /*
     OK All libsw services initilized now
     Initialize switch hw driver object
  */

  r = (*sw)->hw->init( &((*sw)->hwdata), &((*sw)->prm) );

exit:
  /* Exit */
  if (r) {
    libsw_free(*sw);
    (*sw) = 0;
  }

  return r;
}

int libsw_free(libsw_t *lsw)
{
  LIBSWDEBUG("");
  if (lsw) {

    if (LIBSFP_HANDLE(lsw))
      libsfp_free(LIBSFP_HANDLE(lsw));

    if (lsw->dllhandle) {
      if (HW_HANDLE(lsw))
        lsw->hw->free(HW_HANDLE(lsw));
      dlclose(lsw->dllhandle);
    }
    if ( lsw->mac_db )
      lsw->mac_db->close( lsw->mac_db );
    if ( lsw->dbapi )
      lsw->dbapi->close( lsw->dbapi );
    if (lsw->hw)
      free(lsw->hw);
    debug_free(&lsw->log);
    free(lsw);
  }
  return LIBSW_OK;
}

int libsw_connect_hardware(libsw_t *lsw)
{
  return lsw->hw->connect_hardware(HW_HANDLE(lsw));
}

int libsw_init_hardware(libsw_t *lsw)
{
  return lsw->hw->init_hardware_common(HW_HANDLE(lsw));
}

int libsw_reset_hardware(libsw_t *lsw, uint16_t level)
{
  return lsw->hw->reset_hardware(HW_HANDLE(lsw), level);
}

int libsw_get_hw_info(libsw_t *lsw, hw_info_t *info)
{
  return lsw->hw->get_hw_info(HW_HANDLE(lsw), info);
}

/* Object interface */

libsw_obj_t *libsw_get_obj_by_name(libsw_t *lsw,libsw_objid_t id, const char *name)
{
  return lsw->hw->get_obj_by_name(HW_HANDLE(lsw), id, name);
}

libsw_obj_t *libsw_get_obj_by_num(libsw_t *lsw, libsw_objid_t id, uint16_t num)
{
  return lsw->hw->get_obj_by_num(HW_HANDLE(lsw), id, num);
}

libsw_obj_t *libsw_get_obj_by_index(libsw_t *lsw, libsw_objid_t id, uint16_t idx)
{
  return lsw->hw->get_obj_by_idx(HW_HANDLE(lsw), id, idx);
}

libsw_obj_t *libsw_cli_get_obj_by_name(libsw_t *lsw,libsw_objid_t id, const char *name)
{
  uint16_t num;
  LIBSWDEBUGL(1, "id: %u n:'%s'", id, name);

  if (!name[0])
    return 0;

  if (isalpha(name[0]))
    return libsw_get_obj_by_name(lsw, id, name);
  else {
    if (!libsw_s2n10(name, &num))
      return libsw_get_obj_by_num(lsw, id, num);
  }

  return 0;
}

uint16_t libsw_get_obj_index(libsw_t *lsw, const libsw_obj_t *obj)
{
  return lsw->hw->get_obj_index(HW_HANDLE(lsw), obj);
}

uint16_t libsw_get_obj_count(libsw_t *lsw, libsw_objid_t id)
{
  return lsw->hw->get_obj_count(HW_HANDLE(lsw), id);
}

int libsw_set_obj_name(libsw_t *lsw, libsw_obj_t* obj, const char *name)
{
  const char *s = name;

  /* name may be NULL for default */
  if (name) {

    if (!name[0])
      return LIBSW_ERROR;

    if (isdigit(name[0]))
      return LIBSW_ERROR;

    while (*s) {
      if (! (isdigit(*s) || isalpha(*s)) ) {
        return LIBSW_ERROR;
      }
      s++;
    }
  }

  return lsw->hw->set_obj_name(HW_HANDLE(lsw), obj, name);
}

typedef struct {
  libsw_objid_t id;
  char *name;
} objid_item_t;

objid_item_t objid_item_tbl[]=
{
  {LIBSW_OBJID_IFACE, "iface"},
  {LIBSW_OBJID_VLAN, "vlan"},
  {LIBSW_OBJID_LAG, "lag"},
  {LIBSW_OBJID_VMT, "vmt"},
  {LIBSW_OBJID_VIFACE, "viface"}
};

libsw_objid_t libsw_cli_get_objid(libsw_t *lsw, const char *name)
{
  uint8_t i;
  for (i=0; i<(sizeof(objid_item_tbl)/sizeof(objid_item_t)); ++i)
    if (!strcmp(name, objid_item_tbl[i].name))
      return objid_item_tbl[i].id;
  return 0;
}

int libsw_get_obj_stat(libsw_t *lsw, const libsw_obj_t *obj, stat_item_t *i, uint16_t *cnt)
{
  return lsw->hw->get_obj_stat(HW_HANDLE(lsw), obj, i, cnt);
}

int libsw_clear_obj_stat(libsw_t *lsw, const libsw_obj_t *obj)
{
  return lsw->hw->clear_obj_stat(HW_HANDLE(lsw), obj);
}

/* Iterators */
libsw_obj_itr_t *libsw_create_itr(libsw_t *lsw)
{
  libsw_obj_itr_t *it = malloc(sizeof(libsw_obj_itr_t));
  if (it)
    memset(it, 0, sizeof(libsw_obj_itr_t));
  return it;
}

void libsw_init_itr(libsw_t *lsw, libsw_obj_itr_t *it, libsw_objid_t id)
{
  memset(it, 0, sizeof(libsw_obj_itr_t));
  it->id = id;
  it->curnum = 0;  
  it->inc = 1;
  it->lastnum = (libsw_get_obj_count(lsw, id)-1);
}

void  libsw_free_itr(libsw_t *lsw, libsw_obj_itr_t *it)
{
  if (it) {
    free(it);
    it = NULL;
  }
}

int libsw_cli_add_itr_range(libsw_t *lsw, libsw_obj_itr_t *it, const char *range)
{
  it->type |= LIBSW_ITERATOR_COND_RANGE;

  char fname[LIBSW_OBJNAMESIZE];
  char *c;
  libsw_obj_t *obj;

  LIBSWDEBUG("r:'%s' id:%u", range, it->id);  


  c = index(range, '-');

  if (c) {
   memcpy(fname, range, c-range);
   fname[(c-range)] = 0;
  } else
   strncpy(fname, range, LIBSW_OBJNAMESIZE);

  obj = libsw_cli_get_obj_by_name(lsw, it->id, fname);
  if (!obj)
    return 1;

  it->curnum = obj->num;

  if (c) {
    obj = libsw_cli_get_obj_by_name(lsw, it->id, c+1);
    if (!obj)
      return 1;
    it->lastnum = obj->num;
  } else
    it->lastnum = it->curnum;

  it->inc = (it->curnum <= it->lastnum)?1:-1;

  LIBSWDEBUG("fn: %u ln: %u in %i",
             it->curnum,
             it->lastnum, it->inc);
  return 0;
}

int libsw_add_itr_type(libsw_t *lsw, libsw_obj_itr_t *it, uint8_t type)
{
  uint8_t i;
  it->type |= LIBSW_ITERATOR_COND_TYPE;

  for (i=0; i<sizeof(it->types); ++i)
    if (!it->types[i])
      break;

  if (i>=sizeof(it->types))
    return -1;

  if (i<(sizeof(it->types)-1)) {
    it->types[i]=type;
    it->types[i+1]=0;
  }

  return 0;
}

int libsw_set_itr_types(libsw_t *lsw, libsw_obj_itr_t *it, uint8_t *types)
{
  uint8_t i;
  it->type |= LIBSW_ITERATOR_COND_TYPE;

  it->types[0] = 0;
  if (types)
    for (i=0; i<sizeof(it->types); ++i) {
      it->types[i] = types[i];
      if (!types[i])
        break;
    }

  return 0;
}

int libsw_add_itr_flags(libsw_t *lsw, libsw_obj_itr_t *it, uint32_t flags, uint32_t nflags)
{
  it->type |= LIBSW_ITERATOR_COND_FLAGS;
  it->flags |= flags;
  it->nflags |= nflags;

  return 0;
}


int libsw_check_cond_obj(libsw_t *lsw, libsw_obj_itr_t *it, libsw_obj_t *obj)
{
  char found = 0;

  if (it->type & LIBSW_ITERATOR_COND_TYPE) {

    int i;
    for (i=0; i<8; ++i) {
      if (!it->types[i])
        break;
      LIBSWDEBUGL(4,"check type: %02x", it->types[i]);
      if ((it->types[i])==obj->type) {
        found=1;
        break;
      }
    }
    if (!found) {
      LIBSWDEBUGL(4,"wrong type");
      return LIBSW_ERROR;
    }
  }

  if (it->type & LIBSW_ITERATOR_COND_FLAGS) {

    uint32_t oflags;
    if (lsw->hw->get_obj_flags(HW_HANDLE(lsw), obj, &oflags))
      return LIBSW_ERROR;

    if (((it->flags) & (oflags)) != it->flags) {
      LIBSWDEBUGL(4,"wrong flags: it->flags %x, oflags %x", it->flags, oflags);
      return LIBSW_ERROR;
    }
    if ((it->nflags & oflags)) {
      LIBSWDEBUGL(4,"wrong nflags: it->nflags %x, oflags %x", it->nflags, oflags);
      return LIBSW_ERROR;
    }
  }

  return 0;
}

libsw_obj_t *libsw_get_next_obj(libsw_t *lsw, libsw_obj_itr_t *it)
{
  libsw_obj_t *obj=0;

  LIBSWDEBUGL(3, "cn:%u", it->curnum);

  while (1) {

    if (it->inc > 0) {
      if (it->curnum > it->lastnum)
        return 0;
    } else {
      if (it->curnum < it->lastnum)
        return 0;
    }

    LIBSWDEBUGL(4, "cn:%u", it->curnum);
    if (it->type & LIBSW_ITERATOR_COND_RANGE)
      obj = lsw->hw->get_obj_by_num(HW_HANDLE(lsw), it->id, it->curnum);
    else
      obj = lsw->hw->get_obj_by_idx(HW_HANDLE(lsw), it->id, it->curnum);

    it->curnum += it->inc;

    if (!libsw_check_cond_obj(lsw, it, obj)) {
      LIBSWDEBUGL(3,"found obj:%p cn:%i", obj, it->curnum);
      return obj;
    }

  }

}

/*
uint16_t libsw_get_iface_count(libsw_t *lsw)
{    
  return lsw->hw->get_obj_count(HW_HANDLE(lsw), LIBSW_OBJID_IFACE);
}
*/

#if 0
static int libsw_split_name(char *name, char *str, int *num)
{
  int len, r, i;
  uint16_t n;

  len = strlen(name);

  /* find number */
  r=-1;
  for ( i=0; i< len; ++i)
    if (isdigit(name[i])) {
      r=i;
      break;
    }

  if (r<0)
    return -1;

  /* store alpha part of name */
  if (str) {
    if (r!=0)
      memcpy(str, name, r);
    else
      str[0]=0;
  }

  /* store number part */
  if (libsw_s2n10(&name[r], &n))
    return -1;

  (*num) = n;

  return 0;
}
#endif

int libsw_get_obj_status(libsw_t *lsw, const libsw_obj_t *obj, iface_status_t *ifsts)
{
  return lsw->hw->get_obj_status(HW_HANDLE(lsw), obj, ifsts);
}

int libsw_get_iface_status(libsw_t *lsw, const iface_desc_t *iface, iface_status_t *ifsts)
{
  return libsw_get_obj_status(lsw, (libsw_obj_t*)iface, ifsts);
}

int libsw_set_obj_enable(libsw_t *lsw, const libsw_obj_t *obj, uint8_t enable)
{
  CHECK_OBJ_PERM(obj, LIBSW_OPR_ENABLE);
  return lsw->hw->set_obj_enable(HW_HANDLE(lsw), obj, enable);
}

int libsw_set_iface_enable(libsw_t *lsw, const iface_desc_t *iface, uint8_t enable)
{
  return libsw_set_obj_enable(lsw, (libsw_obj_t*)iface, enable);
}

int libsw_set_iface_loopback(libsw_t *lsw, const iface_desc_t *iface, uint8_t enable)
{
  CHECK_OBJ_PERM(iface, LIBSW_OPR_OTHER);
  return lsw->hw->set_iface_loopback(HW_HANDLE(lsw), iface, enable);
}

int libsw_get_iface_portsecurity(libsw_t *lsw, const iface_desc_t *iface, iface_portsec_t * value)
{  
  return lsw->hw->get_iface_portsecurity(HW_HANDLE(lsw), iface, value);
}

int libsw_set_iface_portsecurity(libsw_t *lsw, const iface_desc_t *iface, uint8_t value)
{
  CHECK_OBJ_PERM(iface, LIBSW_OPR_OTHER);
  return lsw->hw->set_iface_portsecurity(HW_HANDLE(lsw), iface, value);
}

int libsw_get_obj_dynamic_learning(libsw_t *lsw, const libsw_obj_t *obj, uint8_t *enable)
{
  return lsw->hw->get_obj_dynamic_learning(HW_HANDLE(lsw), obj, enable);
}

int libsw_get_iface_dynamic_learning(libsw_t *lsw, const iface_desc_t *iface, uint8_t *enable)
{
  return libsw_get_obj_dynamic_learning(lsw, (libsw_obj_t*)iface, enable);
}

int libsw_set_obj_dynamic_learning(libsw_t *lsw, const libsw_obj_t *obj, uint8_t enable)
{
  CHECK_OBJ_PERM(obj, LIBSW_OPR_OTHER);
  return lsw->hw->set_obj_dynamic_learning(HW_HANDLE(lsw), obj, enable);
}

int libsw_set_iface_dynamic_learning(libsw_t *lsw, const iface_desc_t *iface, uint8_t enable)
{
  return libsw_set_obj_dynamic_learning(lsw, (libsw_obj_t*)iface, enable);
}

int libsw_get_obj_defaultvlan(libsw_t *lsw, const libsw_obj_t *obj, vlan_desc_t **vlan)
{
  return lsw->hw->get_obj_defaultvlan(HW_HANDLE(lsw), obj, vlan);
}

int libsw_get_iface_defaultvlan(libsw_t *lsw, const iface_desc_t *iface, vlan_desc_t **vlan)
{
  return libsw_get_obj_defaultvlan(lsw, (libsw_obj_t*)iface, vlan);
}

int libsw_set_obj_defaultvlan(libsw_t *lsw, const libsw_obj_t *obj, vlan_desc_t * vlan)
{
  CHECK_OBJ_PERM(obj, LIBSW_OPR_OTHER);
  /*CHECK_OBJ_PERM(vlan, LIBSW_OPR_VLAN_MEMBERSHIP);*/
  return lsw->hw->set_obj_defaultvlan(HW_HANDLE(lsw), obj, vlan);
}

int libsw_set_iface_defaultvlan(libsw_t *lsw, const iface_desc_t *iface, vlan_desc_t * vlan)
{
  return libsw_set_obj_defaultvlan(lsw, (libsw_obj_t*)iface, vlan);
}

int libsw_add_obj_vlan(libsw_t *lsw, const libsw_obj_t* obj, iface_vlan_t *v)
{
  CHECK_OBJ_PERM(obj, LIBSW_OPR_VLAN_MEMBERSHIP);
  CHECK_OBJ_PERM(v->vlan, LIBSW_OPR_VLAN_MEMBERSHIP);
  return lsw->hw->add_obj_vlan(HW_HANDLE(lsw), obj, v);
}

int libsw_add_iface_vlan(libsw_t *lsw, const iface_desc_t *iface, iface_vlan_t *v)
{
  return libsw_add_obj_vlan(lsw, (const libsw_obj_t*)iface, v);
}

int libsw_del_obj_vlan(libsw_t *lsw, const libsw_obj_t* obj, vlan_desc_t *v)
{
  CHECK_OBJ_PERM(obj, LIBSW_OPR_VLAN_MEMBERSHIP);
  CHECK_OBJ_PERM(v, LIBSW_OPR_VLAN_MEMBERSHIP);
  return lsw->hw->del_obj_vlan(HW_HANDLE(lsw), obj, v);
}

int libsw_del_iface_vlan(libsw_t *lsw, const iface_desc_t *iface, vlan_desc_t *v)
{
  return libsw_del_obj_vlan(lsw, (const libsw_obj_t*)iface, v);
}

int libsw_get_obj_vlans(libsw_t *lsw, const libsw_obj_t *obj,
                        iface_vlan_t *vlans, uint16_t *count)
{
  return lsw->hw->get_obj_vlans(HW_HANDLE(lsw), obj, vlans, count);
}

int libsw_get_iface_vlans(libsw_t *lsw, const iface_desc_t *iface,
                          iface_vlan_t *vlans, uint16_t *count)
{
  return libsw_get_obj_vlans(lsw, (libsw_obj_t*)iface, vlans, count);
}


/**
 * Fills up the given \c members array according to the
 * given \c vlan actual membership information optionally
 * filtering the members by object type.
 *
 * @param lsw libswitch handle
 * @param vlan vlan to get membership information about
 * @param members the members array to fill up
 * @param count the size of the \c members array
 * @param filter object type filter
 *
 * @return 0 on success, non-zero on fail
 */
int libsw_get_vlan_members( libsw_t *lsw, const vlan_desc_t *vlan,
                            vlan_member_t *members, uint16_t *count,
                            libsw_vlan_member_filter_t filter )
{
  return lsw->hw->get_vlan_members( HW_HANDLE( lsw ), vlan, members,
                                    count, filter );
}

stat_item_t *libsw_get_stat_item(stat_item_t *sitems, uint16_t cnt, stat_item_id_t id)
{
  uint16_t i;
  for (i=0; i<cnt; ++i)
    if (sitems[i].id==id)
      return &sitems[i];
  return 0;
}

uint16_t libsw_get_update_time(libsw_t *lsw)
{
  return lsw->hw->get_update_time(HW_HANDLE(lsw));
}

int libsw_update(libsw_t *lsw)
{
  return lsw->hw->update(HW_HANDLE(lsw));
}

int libsw_get_obj_speed_mode(libsw_t *lsw, const libsw_obj_t *obj, speed_mode_t *mode)
{
  return lsw->hw->get_obj_speed_mode(HW_HANDLE(lsw), obj, mode);
}

int libsw_get_iface_speed_mode(libsw_t *lsw, const iface_desc_t *iface, speed_mode_t *mode)
{
  return libsw_get_obj_speed_mode(lsw, (libsw_obj_t*)iface, mode);
}

int libsw_set_iface_speed_mode(libsw_t *lsw, const iface_desc_t *iface, speed_mode_t *mode)
{
  CHECK_OBJ_PERM(iface, LIBSW_OPR_OTHER);
  return lsw->hw->set_iface_speed_mode(HW_HANDLE(lsw), iface, mode);
}

int libsw_get_iface_lag_state(libsw_t *lsw, const iface_desc_t *iface, uint8_t *enable)
{
  return lsw->hw->get_iface_lag_state(HW_HANDLE(lsw), iface, enable);
}

int libsw_set_iface_lag_state(libsw_t *lsw, const iface_desc_t *iface, uint8_t enable)
{
  CHECK_OBJ_PERM(iface, LIBSW_OPR_OTHER);
  return lsw->hw->set_iface_lag_state(HW_HANDLE(lsw), iface, enable);
}

int libsw_get_iface_lag(libsw_t *lsw, const iface_desc_t *iface, lag_desc_t **lag)
{    
  return lsw->hw->get_iface_lag(HW_HANDLE(lsw), iface, lag);
}

int libsw_set_iface_lag(libsw_t *lsw, const iface_desc_t *iface, const lag_desc_t *lag)
{
  CHECK_OBJ_PERM(iface, LIBSW_OPR_OTHER);
  CHECK_OBJ_PERM(lag, LIBSW_OPR_OTHER);
  return lsw->hw->set_iface_lag(HW_HANDLE(lsw), iface, lag);
}

int libsw_set_lag_distribution(libsw_t *lsw, const lag_desc_t * lag, uint16_t mode)
{
  CHECK_OBJ_PERM(lag, LIBSW_OPR_OTHER);
  return lsw->hw->set_lag_distribution(HW_HANDLE(lsw), lag, mode);
}

int libsw_get_lag_distribution(libsw_t *lsw, const lag_desc_t * lag, uint16_t *mode)
{
  return lsw->hw->get_lag_distribution(HW_HANDLE(lsw), lag, mode);
}

int libsw_get_lag_members(libsw_t *lsw, const lag_desc_t * lag, iface_desc_t *ifs[], uint16_t *count)
{
  return lsw->hw->get_lag_members(HW_HANDLE(lsw), lag, ifs, count);
}

int libsw_set_portmirror(libsw_t *lsw, const iface_desc_t *from, const iface_desc_t *to)
{
  CHECK_OBJ_PERM(from, LIBSW_OPR_OTHER);
  return lsw->hw->set_portmirror(HW_HANDLE(lsw), from, to);
}

int libsw_get_portmirror(libsw_t *lsw, const iface_desc_t *from, iface_desc_t **to )
{    
  return lsw->hw->get_portmirror(HW_HANDLE(lsw), from, to);
}

int libsw_set_vlanmirror(libsw_t *lsw, const vlan_desc_t *from, const iface_desc_t *to)
{
  CHECK_OBJ_PERM(from, LIBSW_OPR_OTHER);
  return lsw->hw->set_vlanmirror(HW_HANDLE(lsw), from, to);
}

int libsw_get_vlanmirror(libsw_t *lsw, const vlan_desc_t *from, iface_desc_t **to )
{    
  return lsw->hw->get_vlanmirror(HW_HANDLE(lsw), from, to);
}

static float get_mW2dbm(float value)
{
  float v;

  /* Input value is always not less than 0, we check it above*/
  if ( value > 0.0 ) {
   v = 10.0f * (float)log10(value);
   v = ((float)round(v*10.0f)/10.0f);
  } else {
   v = -41.0f;
  }

  return v;
}


int libsw_get_sfp_info(libsw_t *lsw, const iface_desc_t *iface, sfp_info_t *info)
{
  uint8_t sts;
  sfp_brief_info_t *d;

  LIBSWDEBUG("iface %s", iface->name);

  if ((iface->type != LIBSW_IFTYPE_SFPP) &&
      (iface->type != LIBSW_IFTYPE_SFP))
    return LIBSW_NOT_SUPPORTED;

  if (lsw->hw->get_sfp_status(HW_HANDLE(lsw), iface, &sts))
    return LIBSW_ERROR;

  if (!sts)
    return LIBSW_NOT_INSTALLED;

  d = malloc(sizeof(sfp_brief_info_t));

  if (!d)
    return LIBSW_ERROR;

  lsw->sfp.iface = iface;

  if (libsfp_readinfo_brief(LIBSFP_HANDLE(lsw), d)) {
    free(d);
    return LIBSW_ERROR;
  }

  strcpy(info->vendor, d->vendor);
  strcpy(info->partnum, d->partnum);
  info->spmode =  d->spmode;
  info->rxpower = d->rxpower;
  info->txpower = d->txpower;

  info->rxpower = (d->rxpower>=0.0f)?get_mW2dbm(d->rxpower):-666.0f;
  info->txpower = (d->txpower>=0.0f)?get_mW2dbm(d->txpower):-666.0f;
  info->bitrate = d->bitrate;

  free(d);

  return 0;
}

int libsw_show_sfp_details(libsw_t *lsw, const iface_desc_t *iface, libsw_sfpdetails_t * prms)
{
  uint8_t sts;
  LIBSWDEBUG("iface %s", iface->name);

  if ((iface->type != LIBSW_IFTYPE_SFPP) &&
      (iface->type != LIBSW_IFTYPE_SFP))
    return LIBSW_NOT_SUPPORTED;

  if (lsw->hw->get_sfp_status(HW_HANDLE(lsw), iface, &sts))
    return LIBSW_ERROR;

  if (!sts)
    return LIBSW_NOT_INSTALLED;

  fprintf((prms->file)?prms->file:stdout, "\nSFP module information (%s):\n", iface->name);
  lsw->sfp.iface = iface;
  libsfp_set_outfile(LIBSFP_HANDLE(lsw), (prms->file)?prms->file:stdout);

   if (prms->flags & LIBSW_SFP_DETAILS_FULL)
     libsfp_set_flags(LIBSFP_HANDLE(lsw),
                      LIBSFP_FLAGS_PRINT_BITOPTIONS |
                      LIBSFP_FLAGS_PRINT_CALIBRATIONS |
                      LIBSFP_FLAGS_PRINT_THRESHOLDS |
                      LIBSFP_FLAGS_PRINT_UNKNOWN |
                      LIBSFP_FLAGS_LONGOPT);
   else
     libsfp_set_flags(LIBSFP_HANDLE(lsw), LIBSFP_FLAGS_PRINT_LASERAUTO);


   if (libsfp_showinfo(LIBSFP_HANDLE(lsw)))
      return LIBSW_ERROR;

  return 0;
}

int libsw_get_sfp_speed_mode(libsw_t *lsw, const iface_desc_t *iface, speed_mode_t *smode)
{
  uint32_t spmode;
  uint8_t da;

  smode->speed = 0;
  smode->flags = 0;

  lsw->sfp.iface = iface;
  if (libsfp_get_speed_mode(LIBSFP_HANDLE(lsw), &spmode))
    return LIBSW_ERROR;

  if (libsfp_is_directattach(LIBSFP_HANDLE(lsw), &da))
    return LIBSW_ERROR;

  LIBSWDEBUG("spmode: %u da: %u ", spmode, (uint16_t)da);

  smode->speed = spmode;
  smode->rs0 = LIBSW_IFACE_SPEED_MODE_RSPIN_DEFAULT;
  smode->rs1 = LIBSW_IFACE_SPEED_MODE_RSPIN_DEFAULT;  
  smode->passive = da;

  return 0;
}

int libsw_handle_sfp_after_install(libsw_t *lsw, const iface_desc_t *iface)
{
  return lsw->hw->handle_after_sfp_install(HW_HANDLE(lsw), iface);
}

int libsw_get_mac_table(libsw_t *lsw, mac_table_t *mtable, uint32_t *count)
{
  return lsw->hw->get_mac_table(HW_HANDLE(lsw), mtable, count);
}

int libsw_add_mac_table_entry(libsw_t *lsw, const mac_table_t *mtable)
{
  CHECK_OBJ_PERM(0, LIBSW_OPR_MACTABLE);
  return lsw->hw->add_mac_table_entry(HW_HANDLE(lsw), mtable);
}

int libsw_del_mac_table_entry(libsw_t *lsw, const mac_table_t *mtable)
{
  CHECK_OBJ_PERM(0, LIBSW_OPR_MACTABLE);
  return lsw->hw->del_mac_table_entry(HW_HANDLE(lsw), mtable);
}

/*---------------*/

int libsw_set_iface_vmt(libsw_t *lsw, const iface_desc_t *iface, vmt_desc_t *vmt)
{
  CHECK_OBJ_PERM(iface, LIBSW_OPR_OTHER);
  CHECK_OBJ_PERM(vmt, LIBSW_OPR_OTHER);
  return lsw->hw->set_iface_vmt(HW_HANDLE(lsw), iface, vmt);
}

int libsw_get_iface_vmt(libsw_t *lsw, const iface_desc_t *iface, vmt_desc_t **vmt)
{  
  return lsw->hw->get_iface_vmt(HW_HANDLE(lsw), iface, vmt);
}

int libsw_get_vmt(libsw_t *lsw, const vmt_desc_t *vmt, vmt_item_t *vmt_head, uint16_t *count)
{
  return lsw->hw->get_vmt(HW_HANDLE(lsw), vmt, vmt_head, count);
}

int libsw_add_vmt_item(libsw_t *lsw, const vmt_desc_t *vmt, vmt_item_t *item)
{
  CHECK_OBJ_PERM(0, LIBSW_OPR_OTHER);
  return lsw->hw->add_vmt_item(HW_HANDLE(lsw), vmt, item);
}

int libsw_del_vmt_item(libsw_t *lsw, const vmt_desc_t *vmt, vmt_item_t *item)
{
  CHECK_OBJ_PERM(0, LIBSW_OPR_OTHER);
  return lsw->hw->del_vmt_item(HW_HANDLE(lsw), vmt, item);
}

int libsw_set_stp(libsw_t *lsw, uint8_t en)
{
  CHECK_OBJ_PERM(0, LIBSW_OPR_STP);
  return lsw->hw->set_stp(HW_HANDLE(lsw), en);
}

int libsw_get_stp(libsw_t *lsw, uint8_t *en)
{
  return lsw->hw->get_stp(HW_HANDLE(lsw), en);
}

int libsw_set_stp_mode(libsw_t *lsw, stp_mode_t stpmode)
{
  CHECK_OBJ_PERM(0, LIBSW_OPR_STP);
  return lsw->hw->set_stp_mode(HW_HANDLE(lsw), stpmode);
}

int libsw_get_stp_mode(libsw_t *lsw, stp_mode_t *stpmode)
{
  return lsw->hw->get_stp_mode(HW_HANDLE(lsw), stpmode);
}

int libsw_set_obj_stp(libsw_t *lsw, const libsw_obj_t* obj, const vlan_desc_t *vlan, uint8_t enable)
{
  CHECK_OBJ_PERM(obj, LIBSW_OPR_STP);
  return lsw->hw->set_obj_stp(HW_HANDLE(lsw), obj, vlan, enable);
}

int libsw_set_iface_stp(libsw_t *lsw, const iface_desc_t *iface, const vlan_desc_t *vlan, uint8_t enable)
{    
  return libsw_set_obj_stp(lsw, (const libsw_obj_t*)iface, vlan, enable);
}

int libsw_get_obj_stp(libsw_t *lsw, const libsw_obj_t* obj, const vlan_desc_t *vlan, uint8_t *en)
{    
  return lsw->hw->get_obj_stp(HW_HANDLE(lsw), obj, vlan, en);
}

int libsw_get_iface_stp(libsw_t *lsw, const iface_desc_t *iface, const vlan_desc_t *vlan, uint8_t *en)
{
  return libsw_get_obj_stp(lsw, (const libsw_obj_t*)iface, vlan, en);
}

int libsw_set_obj_stp_state(libsw_t *lsw, const libsw_obj_t* obj,
                            vlan_desc_t *vlan, stp_state_t stpstate)
{
  CHECK_OBJ_PERM(obj, LIBSW_OPR_STP);
  return lsw->hw->set_obj_stp_state(HW_HANDLE(lsw), obj, vlan, stpstate);
}

int libsw_set_iface_stp_state(libsw_t *lsw, const iface_desc_t *iface,
                              vlan_desc_t *vlan, stp_state_t stpstate)
{
  return libsw_set_obj_stp_state(lsw, (const libsw_obj_t*)iface, vlan, stpstate);
}

int libsw_get_obj_stp_state(libsw_t *lsw, const libsw_obj_t* obj, vlan_desc_t *vlan, stp_state_t *stpstate)
{
  return lsw->hw->get_obj_stp_state(HW_HANDLE(lsw), obj, vlan, stpstate);
}

int libsw_get_iface_stp_state(libsw_t *lsw, const iface_desc_t *iface, vlan_desc_t *vlan, stp_state_t *stpstate)
{
  return libsw_get_obj_stp_state(lsw, (const libsw_obj_t*)iface, vlan, stpstate);
}

int libsw_set_stp_params(libsw_t *lsw, const vlan_desc_t *vlan, const libsw_stp_params_t * prm)
{
  CHECK_OBJ_PERM(vlan, LIBSW_OPR_STP);
  return lsw->hw->set_stp_params(HW_HANDLE(lsw), vlan, prm);
}

int libsw_get_stp_params(libsw_t *lsw, const vlan_desc_t *vlan, libsw_stp_params_t * prm)
{
  return lsw->hw->get_stp_params(HW_HANDLE(lsw), vlan, prm);
}

int libsw_set_obj_stp_params(libsw_t *lsw, const libsw_obj_t* obj,
                               const vlan_desc_t *vlan,
                               const libsw_iface_stp_params_t * prm)
{
  CHECK_OBJ_PERM(obj, LIBSW_OPR_STP);
  CHECK_OBJ_PERM(vlan, LIBSW_OPR_STP);
  return lsw->hw->set_obj_stp_params(HW_HANDLE(lsw), obj, vlan, prm);
}

int libsw_set_iface_stp_params(libsw_t *lsw, const iface_desc_t *iface,
                               const vlan_desc_t *vlan,
                               const libsw_iface_stp_params_t * prm)
{
  return libsw_set_obj_stp_params(lsw, (const libsw_obj_t*)iface, vlan, prm);
}


int libsw_get_obj_stp_params(libsw_t *lsw, const libsw_obj_t *obj,
                               const vlan_desc_t *vlan,
                               libsw_iface_stp_params_t * prm)
{
  return lsw->hw->get_obj_stp_params(HW_HANDLE(lsw), obj, vlan, prm);
}

int libsw_get_iface_stp_params(libsw_t *lsw, const iface_desc_t *iface,
                               const vlan_desc_t *vlan,
                               libsw_iface_stp_params_t * prm)
{
  return libsw_get_obj_stp_params(lsw, (const libsw_obj_t*)iface, vlan, prm);
}

int libsw_get_stp_status(libsw_t *lsw, const vlan_desc_t *vlan, libsw_stp_status_t *s)
{
  return lsw->hw->get_stp_status(HW_HANDLE(lsw), vlan, s);
}

int libsw_set_stp_status(libsw_t *lsw, const vlan_desc_t *vlan, const libsw_stp_status_t *s)
{
  CHECK_OBJ_PERM(vlan, LIBSW_OPR_STP);
  return lsw->hw->set_stp_status(HW_HANDLE(lsw), vlan, s);
}

int libsw_get_obj_stp_status(libsw_t *lsw, const libsw_obj_t *obj,
                               const vlan_desc_t *vlan,
                               libsw_iface_stp_status_t *p)
{
  return lsw->hw->get_obj_stp_status(HW_HANDLE(lsw), obj, vlan, p);
}

int libsw_get_iface_stp_status(libsw_t *lsw, const iface_desc_t *iface,
                               const vlan_desc_t *vlan,
                               libsw_iface_stp_status_t *p)
{
  return libsw_get_obj_stp_status(lsw, (const libsw_obj_t*)iface, vlan, p);
}

int libsw_set_obj_stp_status(libsw_t *lsw, const libsw_obj_t *obj,
                               const vlan_desc_t *vlan,
                               const libsw_iface_stp_status_t *p)
{
  CHECK_OBJ_PERM(obj, LIBSW_OPR_STP);
  CHECK_OBJ_PERM(vlan, LIBSW_OPR_STP);
  return lsw->hw->set_obj_stp_status(HW_HANDLE(lsw), obj, vlan, p);
}

int libsw_set_iface_stp_status(libsw_t *lsw, const iface_desc_t *iface,
                                 const vlan_desc_t *vlan,
                                 const libsw_iface_stp_status_t *p)
{
  return libsw_set_obj_stp_status(lsw, (const libsw_obj_t*)iface, vlan, p);
}


int libsw_set_igmp_snooping_enable(libsw_t *lsw, const vlan_desc_t *vlan, uint8_t enable)
{
  CHECK_OBJ_PERM(vlan, LIBSW_OPR_IGMPSNOOP);
  return lsw->hw->set_igmp_snooping_enable(HW_HANDLE(lsw), vlan, enable);
}

int libsw_get_igmp_snooping_enable(libsw_t *lsw, const vlan_desc_t *vlan, uint8_t *enable)
{
  return lsw->hw->get_igmp_snooping_enable(HW_HANDLE(lsw), vlan, enable);
}

int libsw_set_obj_igmp_snooping_enable(libsw_t *lsw, const libsw_obj_t *obj,
                                         const vlan_desc_t *vlan, uint8_t en)
{
  CHECK_OBJ_PERM(obj, LIBSW_OPR_IGMPSNOOP);
  CHECK_OBJ_PERM(vlan, LIBSW_OPR_IGMPSNOOP);
  return lsw->hw->set_obj_igmp_snooping_enable(HW_HANDLE(lsw),
                                                 obj, vlan, en);
}

int libsw_set_iface_igmp_snooping_enable(libsw_t *lsw, const iface_desc_t *iface,
                                         const vlan_desc_t *vlan, uint8_t en)
{
  return libsw_set_obj_igmp_snooping_enable(lsw, (libsw_obj_t*)iface, vlan, en);
}

int libsw_get_obj_igmp_snooping_enable(libsw_t *lsw, const libsw_obj_t *obj,
                                         const vlan_desc_t *vlan, uint8_t *en)
{
  return lsw->hw->get_obj_igmp_snooping_enable(HW_HANDLE(lsw), obj, vlan, en);
}

int libsw_get_iface_igmp_snooping_enable(libsw_t *lsw, const iface_desc_t *iface,
                                         const vlan_desc_t *vlan, uint8_t *en)
{
  return libsw_get_obj_igmp_snooping_enable(lsw, (libsw_obj_t*)iface, vlan, en);
}

int libsw_set_igmp_snooping_params(libsw_t *lsw, const vlan_desc_t *vlan,
                                   const libsw_igmp_snooping_params_t *prms)
{
  return lsw->hw->set_igmp_snooping_params(HW_HANDLE(lsw), vlan, prms);
}

int libsw_get_igmp_snooping_params(libsw_t *lsw, const vlan_desc_t *vlan, 
                                   libsw_igmp_snooping_params_t *prms)
{
  return lsw->hw->get_igmp_snooping_params(HW_HANDLE(lsw), vlan, prms);
}

int libsw_set_obj_igmp_snooping_params(libsw_t *lsw, const libsw_obj_t *obj,
                                         vlan_desc_t *vlan,
                                         const libsw_iface_igmp_snooping_params_t *prm)
{
  return lsw->hw->set_obj_igmp_snooping_params(HW_HANDLE(lsw), obj, vlan, prm);
}

int libsw_set_iface_igmp_snooping_params(libsw_t *lsw, const iface_desc_t *iface,
                                         vlan_desc_t *vlan,
                                         const libsw_iface_igmp_snooping_params_t *prm)
{
  return libsw_set_obj_igmp_snooping_params(lsw, (libsw_obj_t*)iface, vlan, prm);
}


int libsw_get_obj_igmp_snooping_params(libsw_t *lsw, const libsw_obj_t *obj,
                                         const vlan_desc_t *vlan, 
                                         libsw_iface_igmp_snooping_params_t *prm)
{
  return lsw->hw->get_obj_igmp_snooping_params(HW_HANDLE(lsw), obj, vlan, prm);
}

int libsw_get_iface_igmp_snooping_params(libsw_t *lsw, const iface_desc_t *iface,
                                         const vlan_desc_t *vlan,
                                         libsw_iface_igmp_snooping_params_t *prm)
{
  return libsw_get_obj_igmp_snooping_params(lsw, (libsw_obj_t*)iface, vlan, prm);
}

int libsw_acl_init_condition_list(libsw_acl_condition_list_t *rl, uint16_t maxsize)
{
  rl->count = 0;
  rl->condition = malloc(sizeof(libsw_acl_condition_t)*maxsize);
  if (!rl)
    return LIBSW_ERROR;
  memset(rl->condition, 0, sizeof(libsw_acl_condition_t)*maxsize);
  rl->size = maxsize;
  return 0;
}

int libsw_acl_free_condition_list(libsw_acl_condition_list_t *rl)
{
  uint16_t i;

  if (!rl->condition)
    return 1;

  /*printf("%s ");*/
  for (i=0; i<(rl->count); ++i) {
   /* printf("%u con.data = %p ", i, rl->condition[i].data);*/
    if (rl->condition[i].data)
      free(rl->condition[i].data);
  }

  free(rl->condition);
  rl->condition = 0;
  return 0;
}

int libsw_acl_add2condition_list(libsw_acl_condition_list_t *rl, libsw_acl_condition_type_t ruletype,
                                 void *data, uint16_t size, uint8_t flags)
{
  if (rl->count >= rl->size)
    return LIBSW_ERROR;

  rl->condition[rl->count].type = ruletype;
  rl->condition[rl->count].data = malloc(size);
  rl->condition[rl->count].size = size;
  rl->condition[rl->count].flags = flags;

  if (!rl->condition[rl->count].data)
    return LIBSW_ERROR;

  memcpy(rl->condition[rl->count].data, data, size);

  rl->count++;  

  return 0;
}

int libsw_acl_init_rule(libsw_acl_rule_t *rs, uint16_t size)
{
  if (libsw_acl_init_condition_list(&rs->conditions, size))
    return LIBSW_ERROR;
  memset(&rs->action, 0, sizeof(libsw_acl_action_t));

  return 0;
}

int libsw_acl_free_rule(libsw_acl_rule_t *rs)
{
  uint16_t i;

  /*printf("%s rs: %p", rs);*/

  for (i = 0; i < rs->conditions.count; ++i) {
    /*printf("%s %u", i);*/
    libsw_acl_free_condition_list(&(rs[i].conditions));
  }

  if (rs->action.data) {
    free(rs->action.data);
    rs->action.data = 0;
    rs->action.size = 0;
  }

  return 0;
}

int libsw_acl_init_rule_list(libsw_acl_rule_list_t *rs, uint16_t maxsize)
{
  uint16_t i;

  rs->count = 0;
  rs->rules = malloc(sizeof(libsw_acl_rule_t)*maxsize);
  if (!rs)
    return LIBSW_ERROR;
  memset(rs->rules, 0, sizeof(libsw_acl_rule_t)*maxsize);
  rs->size = maxsize;

  for (i=0; i<rs->size; ++i)
    libsw_acl_init_rule(&rs->rules[i], 32);

  return 0;
}

int libsw_acl_free_rule_list(libsw_acl_rule_list_t *rs)
{
  uint16_t i;

  if (!rs->rules)
    return 1;

  for (i=0; i<(rs->count); ++i)
    libsw_acl_free_rule(&rs->rules[i]);
  rs->count = 0;

  free(rs->rules);
  rs->rules = 0;
  return 0;
}

int libsw_acl_add_rule(libsw_t *lsw, libsw_acl_condition_list_t *rl, libsw_acl_action_t *act)
{
  CHECK_OBJ_PERM(0, LIBSW_OPR_ACL);
  return lsw->hw->acl_add_rule(HW_HANDLE(lsw), rl, act);
}

int libsw_acl_del_rule(libsw_t *lsw, libsw_acl_condition_list_t *rl, libsw_acl_action_t *act)
{
  CHECK_OBJ_PERM(0, LIBSW_OPR_ACL);
  return lsw->hw->acl_del_rule(HW_HANDLE(lsw), rl, act);
}

int libsw_acl_del_rule_by_id(libsw_t *lsw, uint16_t ruleid)
{
  CHECK_OBJ_PERM(0, LIBSW_OPR_ACL);
  return lsw->hw->acl_del_rule_by_id(HW_HANDLE(lsw), ruleid);
}

int libsw_acl_get_rules(libsw_t *lsw, libsw_acl_rule_list_t *rsl)
{
  return lsw->hw->acl_get_rules(HW_HANDLE(lsw), rsl);
}

int libsw_acl_clear_all(libsw_t *lsw)
{
  CHECK_OBJ_PERM(0, LIBSW_OPR_ACL);
  return lsw->hw->acl_clear_all(HW_HANDLE(lsw));
}

int libsw_set_mac_aging_time(libsw_t *lsw, uint32_t time)
{
  CHECK_OBJ_PERM(0, LIBSW_OPR_MACTABLE);
  return lsw->hw->set_mac_aging_time(HW_HANDLE(lsw), time);
}

int libsw_get_mac_aging_time(libsw_t *lsw, uint32_t *time)
{
  return lsw->hw->get_mac_aging_time(HW_HANDLE(lsw), time);
}

int libsw_add_ip_iface(libsw_t *lsw, const libsw_ipiface_t *ipif)
{
  CHECK_OBJ_PERM(0, LIBSW_OPR_IPIFACE);
  CHECK_OBJ_PERM(ipif->vlan, LIBSW_OPR_IPIFACE);
  return lsw->hw->change_ip_iface(HW_HANDLE(lsw), ipif, LIBSW_CH_ACTION_ADD);
}

int libsw_del_ip_iface(libsw_t *lsw, const libsw_ipiface_t *ipif)
{
  CHECK_OBJ_PERM(0, LIBSW_OPR_IPIFACE);
  CHECK_OBJ_PERM(ipif->vlan, LIBSW_OPR_IPIFACE);
  return lsw->hw->change_ip_iface(HW_HANDLE(lsw), ipif, LIBSW_CH_ACTION_DEL);
}

int libsw_get_ip_ifaces(libsw_t *lsw, libsw_ipiface_t *ifs, uint16_t *count)
{
  return lsw->hw->get_ip_ifaces(HW_HANDLE(lsw), ifs, count);
}

int libsw_add_ip_route(libsw_t *lsw, const vlan_desc_t *vlan, const libsw_iproute_t *rt)
{
  CHECK_OBJ_PERM(0, LIBSW_OPR_IPROUTE);
  CHECK_OBJ_PERM(vlan, LIBSW_OPR_IPROUTE);
  return lsw->hw->change_ip_route(HW_HANDLE(lsw), vlan, rt, LIBSW_CH_ACTION_ADD);
}

int libsw_del_ip_route(libsw_t *lsw, const vlan_desc_t *vlan, const libsw_iproute_t *rt)
{
  CHECK_OBJ_PERM(0, LIBSW_OPR_IPROUTE);
  CHECK_OBJ_PERM(vlan, LIBSW_OPR_IPROUTE);
  return lsw->hw->change_ip_route(HW_HANDLE(lsw), vlan, rt, LIBSW_CH_ACTION_DEL);
}

int libsw_get_ip_routes(libsw_t *lsw, const vlan_desc_t *vlan, libsw_iproute_t *rts, uint16_t *count)
{
  return lsw->hw->get_ip_routes(HW_HANDLE(lsw), vlan, rts, count);
}

#include <netdb.h>

int libsw_get_l4_proto_num(char *pname, uint8_t *pnum)
{
  struct protoent *p;
  uint16_t n;

  if (!pname)
    return LIBSW_ERROR;

  if (isdigit(pname[0])) {
    if (libsw_s2n10(pname, &n))
      return LIBSW_ERROR;
    (*pnum) = n;
  } else {
    if (!(p=getprotobyname(pname)))
      return LIBSW_ERROR;
    (*pnum) = p->p_proto;
  }
  return 0;
}

int libsw_get_l4_proto_name(uint8_t pnum, char *pname, uint8_t l)
{
  struct protoent *p;
  if (!(p=getprotobynumber(pnum)))
    return LIBSW_ERROR;

  strncpy(pname, p->p_name, l);

  return 0;
}

int libsw_get_l4_serv_port(char *sname, uint16_t *pnum)
{
  struct servent *p;
  uint16_t n;

  if (!sname)
    return LIBSW_ERROR;

  if (isdigit(sname[0])) {
    if (libsw_s2n10(sname, &n))
      return LIBSW_ERROR;
    (*pnum) = n;
  } else {
    if (!(p=getservbyname(sname, 0)))
      return LIBSW_ERROR;
    (*pnum) = ntohs(p->s_port);
  }
  return 0;
}

int libsw_get_l4_serv_name(uint16_t pnum, char *sname, uint8_t l)
{
  struct servent *p;
  if (!(p=getservbyport(htons(pnum), 0)))
    return LIBSW_ERROR;

  strncpy(sname, p->s_name, l);

  return 0;
}

int libsw_open_slow_protocol_fd(libsw_t *lsw, int *fd, libsw_slow_prot_id_t *slp)
{    
  return lsw->hw->open_slow_protocol_fd(HW_HANDLE(lsw), fd, slp);
}

int libsw_read_slow_protocol_packet(libsw_t *lsw, int fd, uint8_t flags, char *buf, size_t *size, libsw_slow_prot_hdr_t *hdr)
{
  return lsw->hw->read_slow_protocol_packet(HW_HANDLE(lsw), fd, flags, buf, size, hdr);
}

/**
 * Prepares the packet to be sent as a slow protocol packet over some
 * media. Performs the necessary data encoding and optionally sets the
 * VLAN tag. The packet data buffer has to be at least 2-bytes larger than
 * the packet length and multiple of 4 bytes.
 *
 * The new packet length is returned in \c len on success.
 * 
 * @param h libsw handle
 * @param obj media object (an interface or a LAG, optional)
 * @param vlan VLAN descriptor (optional)
 * @param data packet data
 * @param len packet length
 * @param bufsize the whole available buffer size
 * 
 * @return 0 on success, non-zero on fail
 */
int libsw_prepare_packet_for_send( libsw_t *lsw,
                                   const libsw_obj_t *obj,
                                   const vlan_desc_t *vlan,
                                   void *data,
                                   size_t *len,
                                   size_t bufsize )
{
  return lsw->hw->prepare_packet_for_send(
                      HW_HANDLE( lsw ), obj, vlan, data, len,
                      bufsize );
}

/**
 * Sends the given data as a slow protocol packet over the given
 * media object \c obj. The data MUST be prepared for sending with
 * a call to the #libsw_prepare_packet_for_send.
 *
 * @param h libsw handle
 * @param obj media object (an interface or a LAG)
 * @param prepared_data prepared data buffer to be sent
 * @param size data buffer size
 *
 * @return 0 on success, non-zero on fail
 */
int libsw_send_packet( libsw_t *lsw,
                       const libsw_obj_t *obj,
                       const void *prepared_data,
                       size_t size )
{
  CHECK_OBJ_PERM( obj, LIBSW_OPR_PACKET );

  return lsw->hw->send_packet( HW_HANDLE( lsw ),
                               obj,
                               prepared_data,
                               size );
}

typedef struct {
  int code;
  char *msg;
} libsw_error_msg_t;

libsw_error_msg_t error_msg_tbl[]={
    {LIBSW_OK, "OK"},
    {LIBSW_ERROR, "Operation failed"},
    {LIBSW_NOT_FOUND, "Object not found"},
    {LIBSW_NOT_SUPPORTED, "Operation not supported"},
    {LIBSW_NOT_ALLOWED, "Operation not allowed"},
  };

char *libsw_strerror(int code)
{
  uint16_t i;
  for (i=0; i<sizeof(error_msg_tbl)/sizeof(libsw_error_msg_t);++i)
    if (error_msg_tbl[i].code==code)
      return error_msg_tbl[i].msg;
  return "Unknown error";
}

/**
 * \brief Hide/unhide object from listing and from group operation
 * \param lsw      - libsw handle
 * \param obj      - object to hide
 * \param hidden   - 0/1 clear/set
 * \return 0 if success, nonzero otherwise
*/
int libsw_set_obj_hidden(libsw_t *lsw, libsw_obj_t* obj, uint8_t hidden)
{
  CHECK_OBJ_PERM(obj, LIBSW_OPR_HIDE);

  return lsw->hw->set_obj_flags(HW_HANDLE(lsw), obj,
                                (hidden)?LIBSW_OBJFLAG_HIDDEN:0,
                                LIBSW_OBJFLAG_HIDDEN);  
}


int libsw_set_management_configuration_ready( libsw_t *lsw, char ready )
{
  return lsw->hw->set_management_configuration_ready(HW_HANDLE(lsw), ready);
}

int libsw_check_jumbo_available( libsw_t *lsw )
{
  return lsw->hw->check_jumbo_available( HW_HANDLE(lsw) );
}

int libsw_set_port_jumbo_frame_length( libsw_t *lsw, const iface_desc_t *iface, uint16_t len)
{
  return lsw->hw->set_port_jumbo_frame_length( HW_HANDLE(lsw), iface, len );
}

int libsw_port_disable_jumbo_frames_support( libsw_t *lsw, const iface_desc_t *iface )
{
  return lsw->hw->disable_jumbo_frames_support( HW_HANDLE(lsw), iface );
}

int libsw_get_iface_frame_length( libsw_t *lsw, const iface_desc_t *iface, uint16_t *len )
{
  return lsw->hw->get_iface_frame_length(HW_HANDLE(lsw), iface, len);
}
